// EpicsProcessVariable.cpp: implementation of the CEpicsProcessVariable class.
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Config.h"
#include "EpicsProcessVariable.h"
#include "Win32Unicode.h"
#include "VariantTools.h"

//////////////////////////////////////////////////////////////////////
// CDoubleValue
//////////////////////////////////////////////////////////////////////

CDoubleValue::CDoubleValue ()
{
	m_precision = 0;
	m_value = NULL;
}

CDoubleValue::~CDoubleValue ()
{
	if (m_value)
		delete [] m_value;
}

BOOL CDoubleValue::Init (CEpicsProcessVariable *epv, unsigned long count)
{
	CGenericValue::Init (epv, count);
	m_value = new double[count];
	if (m_value)
		memset (m_value, 0, sizeof(double)*count);
	m_count = count;

	return m_value != NULL;
}

chtype CDoubleValue::GetRequestType ()
{
	return DBR_DOUBLE;
}

void CDoubleValue::CtrlCallback (struct event_handler_args *args)
{
	struct dbr_ctrl_double *ctrl = (struct dbr_ctrl_double *) args->dbr;
	unsigned long i, count = args->count;
	dbr_double_t *data = &ctrl->value;

	strcpy (m_units, ctrl->units);
	m_precision = ctrl->precision;

	if (count > m_count)
		count = m_count;
	for (i=0; i<count; ++i)
		m_value[i] = data[i];
}

void CDoubleValue::EventCallback (struct event_handler_args *args)
{
	dbr_double_t *data = (dbr_double_t *) args->dbr;
	unsigned long i, count = args->count;

#ifdef _DEBUG
	if ((unsigned long) args->count != m_count)
		ATLTRACE("EventCallback '%s': my count %ld <> CA callback count of %ld\n",
			m_epv->GetName(), m_count, args->count);
#endif

	if (count > m_count)
		count = m_count;
	for (i=0; i<count; ++i)
		m_value[i] = data[i];
}

// The VARIANT will be initialized
// and has to be free'd with VariantClear ()
BOOL CDoubleValue::GetValue (VARIANT *dest)
{
	SAFEARRAYBOUND	bound[1];
	unsigned long i;
	double *arraydata;

	if (! dest)
		return FALSE;

	VariantInit (dest);

	if (m_count == 1)
	{
		dest->vt = VT_R8;
		dest->dblVal = *m_value;

#if TRACELEVEL >= 4
		ATLTRACE ("CDoubleValue '%s':\n", m_epv->GetName ());
		ShowVariant (*dest);
#endif

		return TRUE;
	}

	// I'm not sure which shortcuts would be safe,
	// so let's stick to the elaborate procedure.
	// Attach SAFEARRAY and adjust type:
	bound[0].lLbound = 0;
	bound[0].cElements = m_count;
	dest->parray = SafeArrayCreate(VT_R8, 1, bound);
	if (!dest->parray)
		return FALSE;
	dest->vt = VT_ARRAY | VT_R8;

	// copy values into the SAFEARRAY
	SafeArrayAccessData (dest->parray, (void **) &arraydata);
	for (i=0; i<m_count; ++i)
		arraydata[i] = m_value[i];
	SafeArrayUnaccessData (dest->parray);

#if TRACELEVEL >= 4
	ATLTRACE ("CDoubleValue '%s':\n", m_epv->GetName ());
	ShowVariant (*dest);
#endif

	return TRUE;
}

//////////////////////////////////////////////////////////////////////
// CI16Value
//////////////////////////////////////////////////////////////////////

CI16Value::CI16Value ()
{
	m_value = NULL;
}

CI16Value::~CI16Value ()
{
	if (m_value)
		delete [] m_value;
}

BOOL CI16Value::Init (CEpicsProcessVariable *epv, unsigned long count)
{
	CGenericValue::Init (epv, count);
	m_value = new short[count];
	if (m_value)
		memset (m_value, 0, sizeof(short)*count);
	m_count = count;

	return m_value != NULL;
}

chtype CI16Value::GetRequestType ()
{
	return DBF_SHORT;
}

void CI16Value::CtrlCallback (struct event_handler_args *args)
{
	struct dbr_ctrl_short *ctrl = (struct dbr_ctrl_short *) args->dbr;
	unsigned long i, count = args->count;
	dbr_short_t *data = &ctrl->value;

	strcpy (m_units, ctrl->units);

	if (count > m_count)
		count = m_count;
	for (i=0; i<count; ++i)
		m_value[i] = data[i];
}

void CI16Value::EventCallback (struct event_handler_args *args)
{
	dbr_short_t *data = (dbr_short_t *) args->dbr;
	unsigned long i, count = args->count;

#ifdef _DEBUG
	if ((unsigned long) args->count != m_count)
		ATLTRACE("EventCallback '%s': my count %ld <> CA callback count of %ld\n",
			m_epv->GetName(), m_count, args->count);
#endif

	if (count > m_count)
		count = m_count;
	for (i=0; i<count; ++i)
		m_value[i] = data[i];
}

// The VARIANT will be initialized
// and has to be free'd with VariantClear ()
BOOL CI16Value::GetValue (VARIANT *dest)
{
	SAFEARRAYBOUND	bound[1];
	unsigned long i;
	short *arraydata;

	if (! dest)
		return FALSE;

	VariantInit (dest);

	if (m_count == 1)
	{
		dest->vt = VT_I2;
		dest->iVal = *m_value;

#if TRACELEVEL >= 4
		ATLTRACE("CI16Value::GetValue '%s' %d = 0x%0X\n",
            m_epv->GetName(), (int)*m_value, (int)*m_value);
		ShowVariant (*dest);
#endif

		return TRUE;
	}

	// I'm not sure which shortcuts would be safe,
	// so let's stick to the elaborate procedure.
	// Attach SAFEARRAY and adjust type:
	bound[0].lLbound = 0;
	bound[0].cElements = m_count;
	dest->parray = SafeArrayCreate(VT_I2, 1, bound);
	if (!dest->parray)
		return FALSE;
	dest->vt = VT_ARRAY | VT_I2;

	// copy values into the SAFEARRAY
	SafeArrayAccessData (dest->parray, (void **) &arraydata);
	for (i=0; i<m_count; ++i)
		arraydata[i] = m_value[i];
	SafeArrayUnaccessData (dest->parray);

#if TRACELEVEL >= 4
	ATLTRACE ("CI16Value::GetValue '%s':\n", m_epv->GetName ());
	ShowVariant (*dest);
#endif

	return TRUE;
}

BOOL CStringValue::GetValue(VARIANT *dest)
{
	if (! dest)
		return FALSE;

    LPOLESTR olestr;
    if (AnsiToUnicode(m_value, &olestr) == NOERROR)
    {
	    VariantInit (dest);
        dest->vt = VT_BSTR;
	    dest->bstrVal = SysAllocString(olestr);
        CoTaskMemFree(olestr);
    }
    else
        return FALSE;

#if TRACELEVEL >= 4
	ATLTRACE ("CStringValue '%s':\n", m_epv->GetName ());
	ShowVariant (*dest);
#endif

    return TRUE;
}   

//////////////////////////////////////////////////////////////////////
// CEpicsProcessVariable
//////////////////////////////////////////////////////////////////////

CGenericValue *CEpicsProcessVariable::CreateValue (chtype type, unsigned long count)
{
	CGenericValue *val;

	switch (type)
	{
	case DBF_STRING:
		val = new CStringValue();
		break;
	case DBF_SHORT:
	case DBF_ENUM:
	case DBF_CHAR:
		val = new CI16Value ();
		break;
	default:
		val = new CDoubleValue ();
	}

	if (val && val->Init (this, count))
		return val;

	return NULL;
}


CEpicsProcessVariable::CEpicsProcessVariable()
{
	m_name = NULL;
	m_chid = NULL;
	m_evid = NULL;
	m_connected = FALSE;
	m_value = NULL;
    m_use_put_callback = FALSE;
	m_new_value_func = NULL;
	m_new_value_func_arg = NULL;
}

CEpicsProcessVariable::~CEpicsProcessVariable()
{
    StopMonitor();
    ca_clear_channel(m_chid);
    ATLTRACE("delete CEpicsProcessVariable(%s) in Thread 0x%08X\n",
		    m_name, GetCurrentThreadId());
	if (m_name)
		delete [] m_name;
	if (m_value)
		delete m_value;

#ifdef _DEBUG
	m_name = NULL;
	m_value = NULL;
#endif
}

//	Usual trace of calls:
//
//	(1)
//	setup a connection-handler for  m_pv_name

// The supplied notify routine will be called upon
// successful connection.
//
// Returns NULL on error.
CEpicsProcessVariable *CEpicsProcessVariable::Create (const char *name, EPVNotifyFunc connected, void *arg)
{
	CEpicsProcessVariable	*epv = new CEpicsProcessVariable;
	int res;

	if (! epv)
		return NULL;

	ATLTRACE("CEpicsProcessVariable::Create (%s) in Thread 0x%08X\n",
		name, GetCurrentThreadId());

	epv->m_connection_func = connected;
	epv->m_connection_func_arg = arg;

	epv->m_name = new char[strlen(name)+1];
	if (epv->m_name)
	{
		strcpy (epv->m_name, name);

		res = ca_search_and_connect(epv->m_name, &epv->m_chid, ConnectCallback, epv);
		ca_pend_io (2);
		if (res != ECA_NORMAL)
		{
			delete epv;
			return NULL;
		}
	}

	return epv;
}

//	(2)
//	connected, i.e. we know the native data type & size
//	-> set up a C*Value for this type & get control info
void CEpicsProcessVariable::ConnectCallback (struct connection_handler_args args)
{
	CEpicsProcessVariable *me = (CEpicsProcessVariable *) ca_puser(args.chid);
	int res;

	//	This was an assert which failed some time,
	//	so now we just ignore it
	if (me->m_chid != args.chid)
		return;

	ATLTRACE("CEpicsProcessVariable::ConnectCallback (%s)\n", me->m_name);

	switch (args.op)
	{
	case CA_OP_CONN_UP:
		if (! me->m_value)
			me->m_value = me->CreateValue (ca_get_field_type (me->m_chid), ca_get_element_count(me->m_chid));
		if (!me->m_value)
			return;

		res = ca_array_get_callback (me->m_value->GetControlType(),
#define HACK
		// Problem: CA server returns garbage when asked for CTRL on array.
		// Jeff claims he fixed this in the latest server,
		// but we don't really need all the values here so the HACK
		// might be the safest option for a while.
#ifdef HACK
			1,
#else
			ca_get_element_count(me->m_chid),
#endif
				me->m_chid, CtrlCallback, me);
		ca_flush_io ();
		//	signal of new state will be done in
		//	CtrlCallback when we know the limits, units etc.
		return;
	case CA_OP_CONN_DOWN:
		me->m_connected = FALSE;
		if (me->m_connection_func)
			me->m_connection_func (me->m_connection_func_arg);
		return;
	}
}

//	(3)
//	Received Info, initiate events
void CEpicsProcessVariable::CtrlCallback(struct event_handler_args args)
{
	CEpicsProcessVariable	*me = (CEpicsProcessVariable *) args.usr;

	ATLTRACE("CEpicsProcessVariable::CtrlCallback (%s)\n", me->m_name);

	//	Let C*Value handle the control data (units, limits, etc.)
	//	we received...
	me->m_value->CtrlCallback (&args);
	me->m_connected = TRUE;

	if (me->m_connection_func)
		me->m_connection_func (me->m_connection_func_arg);
}

BOOL CEpicsProcessVariable::SetShortValue(const short *value, unsigned long count)
{
	ATLTRACE("CEpicsProcessVariable::SetShortValue (%s) in Thread 0x%08X\n",
		m_name, GetCurrentThreadId());
    int status;
    if (m_use_put_callback)
        status = ca_array_put_callback(DBF_SHORT, count, m_chid, value, PutCallback, this);
    else
        status = ca_array_put(DBF_SHORT, count, m_chid, value);
	return status == ECA_NORMAL && ca_flush_io() == ECA_NORMAL;
}

BOOL CEpicsProcessVariable::SetLongValue(const long *value, unsigned long count)
{
	ATLTRACE("CEpicsProcessVariable::SetLongValue (%s) in Thread 0x%08X\n",
		m_name, GetCurrentThreadId());
    int status;
    if (m_use_put_callback)
        status = ca_array_put_callback(DBF_LONG, count, m_chid, value, PutCallback, this);
    else
        status = ca_array_put(DBF_LONG, count, m_chid, value);
	return status == ECA_NORMAL && ca_flush_io() == ECA_NORMAL;
}

BOOL CEpicsProcessVariable::SetFloatValue(const float *value, unsigned long count)
{
	ATLTRACE("CEpicsProcessVariable::SetFloatValue (%s) in Thread 0x%08X\n",
		m_name, GetCurrentThreadId());
    int status;
    if (m_use_put_callback)
        status = ca_array_put_callback(DBF_FLOAT, count, m_chid, value, PutCallback, this);
    else
        status = ca_array_put(DBF_FLOAT, count, m_chid, value);
	return status == ECA_NORMAL && ca_flush_io() == ECA_NORMAL;
}

BOOL CEpicsProcessVariable::SetDoubleValue(const double *value, unsigned long count)
{
	ATLTRACE("CEpicsProcessVariable::SetDoubleValue (%s) in Thread 0x%08X\n",
		m_name, GetCurrentThreadId());
    int status;
    if (m_use_put_callback)
        status = ca_array_put_callback(DBF_DOUBLE, count, m_chid, value, PutCallback, this);
    else
        status = ca_array_put(DBF_DOUBLE, count, m_chid, value);
	return status == ECA_NORMAL && ca_flush_io() == ECA_NORMAL;
}

BOOL CEpicsProcessVariable::SetStringValue(const char *value)
{
	ATLTRACE("CEpicsProcessVariable::SetStringValue (%s) in Thread 0x%08X\n",
		m_name, GetCurrentThreadId());
    int status;
    if (m_use_put_callback)
        status = ca_put_callback(DBF_STRING, m_chid, value, PutCallback, this);
    else
        status = ca_put(DBF_STRING, m_chid, value);
	return status == ECA_NORMAL && ca_flush_io() == ECA_NORMAL;
}
// Start monitor for this PV (only once)
void CEpicsProcessVariable::StartMonitor(EPVNotifyFunc func, void *arg)
{
	if (m_evid)
		StopMonitor();

	m_new_value_func = func;
	m_new_value_func_arg = arg;

	ca_add_masked_array_event(m_value->GetRequestType(),
		ca_get_element_count(m_chid), m_chid,
		EventCallback, this, 0.0, 0.0, 0.0,
		&m_evid, DBE_VALUE | DBE_ALARM);
}

void CEpicsProcessVariable::IssueGetCallback()
{
	ATLTRACE("CEpicsProcessVariable::IssueGetCallback(%s)\n", m_name);
	ca_array_get_callback(m_value->GetRequestType(),
		ca_get_element_count(m_chid), m_chid,
		EventCallback, this);
}

void CEpicsProcessVariable::EventCallback(struct event_handler_args args)
{
	CEpicsProcessVariable *me = (CEpicsProcessVariable *) args.usr;

	ATLASSERT(me != NULL);
	// if (args.status != ECA_NORMAL) {
	//			ca_message(args.status));

	//	check that the value received matches the one requested
	//	because PVValue will blindly assume this:
	if (args.type != me->m_value->GetRequestType())
        return;

	ATLTRACE("CEpicsProcessVariable::EventCallback(%s)\n", me->m_name);

	// Allow C*Value to handle the received value
	me->m_value->EventCallback(&args);

	// Notify whoever was interested in this event:
	if (me->m_new_value_func)
		me->m_new_value_func(me->m_new_value_func_arg);
}

void CEpicsProcessVariable::StopMonitor()
{
	if (m_evid)
	{
		ca_clear_event(m_evid);
		m_evid = NULL;
	}
}

void CEpicsProcessVariable::PutCallback(struct event_handler_args args)
{
	CEpicsProcessVariable *me = (CEpicsProcessVariable *) args.usr;

	ATLASSERT(me != NULL);
	// if (args.status != ECA_NORMAL) {
	//			ca_message(args.status));

	ATLTRACE("CEpicsProcessVariable::PutCallback(%s), status=%d\n",
        me->m_name, (int)args.status);
	
	// Notify whoever was interested in this event:
	if (me->m_new_value_func)
		me->m_new_value_func(me->m_new_value_func_arg);
}

