function client_test(name)
%CLIENT_TEST
%
%  client_test(pv_name)
%
%  Example: client_test('snsioc2:vxiRegInB0_00')
%
%  Demo of reading one value via EpicsCAClient.
%  Please look at the source comments.
%
%  kasemir@lanl.gov

% Connect to the "EpicsCAClient" ActiveX Automation Server,
% create a "ProcessVariable" on there:
pv = actxserver('EpicsCAClient.ProcessVariable');

% Set the "name" property of the EpicsCAClient.ProcessVariable
pv.name = name;

% Now the EpicsCAClient program will try to establish a ChannelAccess
% connection to a channel of that name. This takes some time because
% we are going over the network. It could also fail. What do we do?
%  
% The EpicsCAClient.Processvariable sends an ActiveX Event when the connection
% status and/or value changes.
% In Visual Basic we would prepare to catch the "Changed" event and spend our
% time doing something useful until we receive the "Changed" event.
% But Matlab cannot handle events for ActiveX _servers_ (only "controls").
% So all we can do is sit in a busy wait loop until the connection comes up.
% If the PV doesn't exist, we will be stuck in here forever....
wait=0;
while pv.is_connected == 0,
    wait=wait+1;
end

% Ok, we are connected.
% In a 'real' program you would do the connection stuff only once,
% and then spend a lot of time in highly interesting calculations
% based on the value.
% Here we just print the name & value once:
disp(pv.name);
disp(pv.value);

% Ask the EpicsCAServer to remove out ProcessVariable.
% It will disconnect from ChannelAccess, the 'pv' is useless from now on:
delete(pv);




