// ProcessVariable.h : Declaration of the CProcessVariable

#ifndef __PROCESSVARIABLE_H_
#define __PROCESSVARIABLE_H_

#include "resource.h"       // main symbols
#include "EpicsCAServerCP.h"

/////////////////////////////////////////////////////////////////////////////
// CProcessVariable

//	CEpicsPV is the casPV class for this server,
//	CProcessVariable handles the COM interface.
//	Both are linked and can only exist as twins,
//	still they are separated because it's hopefully
//	easier to see the COM/CA differences:

#ifdef _DEBUG
// Yet another single-linked list template,
// but this one is simple enough so that the
// MS debugger can understand it at run time
template <class T>
class SLList
{
public:
	SLList ()
	{
		sl_next = 0;
	}

	void add (T *item)
	{
		item->sl_next = sl_next;
		sl_next = item;
	}

	bool remove (T *item)
	{
		T **current = &sl_next;

		while (*current != item)
			current = & ((*current)->sl_next);

		if (*current)
		{
			*current = (*current)->sl_next;
			return true;
		}

		return false;
	}

protected:
	T *sl_next;
};
#endif


class ATL_NO_VTABLE CProcessVariable : 
#ifdef _DEBUG
	public SLList<CProcessVariable>,
#endif
	public CComObjectRootEx<CComSingleThreadModel>,
	public CComCoClass<CProcessVariable, &CLSID_ProcessVariable>,
	public ISupportErrorInfo,
	public IConnectionPointContainerImpl<CProcessVariable>,
	public IDispatchImpl<IProcessVariable, &IID_IProcessVariable, &LIBID_EpicsCAServer>,
	public IProvideClassInfo2Impl<&CLSID_ProcessVariable, &DIID__IProcessVariableEvents, &LIBID_EpicsCAServer>,
	public CProxy_IProcessVariableEvents< CProcessVariable >
{
public:
	CProcessVariable();
	~CProcessVariable();

DECLARE_REGISTRY_RESOURCEID(IDR_PROCESSVARIABLE)

DECLARE_PROTECT_FINAL_CONSTRUCT()

BEGIN_COM_MAP(CProcessVariable)
	COM_INTERFACE_ENTRY(IProcessVariable)
	COM_INTERFACE_ENTRY(IDispatch)
	COM_INTERFACE_ENTRY(ISupportErrorInfo)
	COM_INTERFACE_ENTRY(IConnectionPointContainer)
	COM_INTERFACE_ENTRY(IProvideClassInfo2)
	COM_INTERFACE_ENTRY(IProvideClassInfo)
	COM_INTERFACE_ENTRY_IMPL(IConnectionPointContainer)
END_COM_MAP()
BEGIN_CONNECTION_POINT_MAP(CProcessVariable)
//CONNECTION_POINT_ENTRY(IID__IProcessVariableEvents)
CONNECTION_POINT_ENTRY(DIID__IProcessVariableEvents)
END_CONNECTION_POINT_MAP()


// ISupportsErrorInfo
	STDMETHOD(InterfaceSupportsErrorInfo)(REFIID riid);

// IProcessVariable
public:
	STDMETHOD(SetValuesAndTimes)(VARIANT values, VARIANT times);
	STDMETHOD(GetEpicsTime)(unsigned long *seconds, unsigned long *nanosecs);
	STDMETHOD(SetEnumValueAndTime)(long value, unsigned long secs, unsigned long nanosecs);
	STDMETHOD(SetEnumValue)(long value);
	STDMETHOD(SetValueAndTime)(VARIANT new_value, unsigned long secs, unsigned long nanosecs);
	STDMETHOD(SetValue)(VARIANT new_value);
	STDMETHOD(put_enum_string)(int index, /*[in]*/ BSTR newVal);
	STDMETHOD(get_deadband)(/*[out, retval]*/ double *pVal);
	STDMETHOD(put_deadband)(/*[in]*/ double newVal);
	STDMETHOD(get_precision)(/*[out, retval]*/ long *pVal);
	STDMETHOD(put_precision)(/*[in]*/ long newVal);
	STDMETHOD(get_control_high)(/*[out, retval]*/ double *pVal);
	STDMETHOD(put_control_high)(/*[in]*/ double newVal);
	STDMETHOD(get_control_low)(/*[out, retval]*/ double *pVal);
	STDMETHOD(put_control_low)(/*[in]*/ double newVal);
	STDMETHOD(get_graphic_high)(/*[out, retval]*/ double *pVal);
	STDMETHOD(put_graphic_high)(/*[in]*/ double newVal);
	STDMETHOD(get_graphic_low)(/*[out, retval]*/ double *pVal);
	STDMETHOD(put_graphic_low)(/*[in]*/ double newVal);
	STDMETHOD(get_alarm_high)(/*[out, retval]*/ double *pVal);
	STDMETHOD(put_alarm_high)(/*[in]*/ double newVal);
	STDMETHOD(get_alarm_low)(/*[out, retval]*/ double *pVal);
	STDMETHOD(put_alarm_low)(/*[in]*/ double newVal);
	STDMETHOD(get_high_warning)(/*[out, retval]*/ double *pVal);
	STDMETHOD(put_high_warning)(/*[in]*/ double newVal);
	STDMETHOD(get_low_warning)(/*[out, retval]*/ double *pVal);
	STDMETHOD(put_low_warning)(/*[in]*/ double newVal);
	STDMETHOD(get_units)(/*[out, retval]*/ BSTR *pVal);
	STDMETHOD(put_units)(/*[in]*/ BSTR newVal);
	STDMETHOD(get_name)(/*[out, retval]*/ BSTR *pVal);
	STDMETHOD(put_name)(/*[in]*/ BSTR newVal);

private:
	// link to EPICS PV:
	class CEpicsPV	*m_epics_pv;

	// internally used to handle SetValue with Array VARIANT
    HRESULT SetValue(VARIANT new_value, struct timespec &time);
	HRESULT SetArrayValue (VARTYPE type, SAFEARRAY *arr, struct timespec &time);
};

#endif //__PROCESSVARIABLE_H_
