// Server.cpp: implementation of the CServer class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Server.h"
#include "EpicsPV.h"

CServer::CServer(size_t hash_table_size)
{
	m_PV_table.init(hash_table_size);
	CEpicsPV::StaticInit (this);
}

CServer::~CServer()
{
	m_PV_table.traverse(CEpicsPV::Unregister);
}

// Tell server to serve this PV
void CServer::Register (CEpicsPV &epv)
{
	m_PV_table.add (epv);

	ATLTRACE(_T("CServer::Register (PV '%s')\n"), epv.getName());
}

// Tell server not to serve this PV any longer
void CServer::Unregister (CEpicsPV &epv)
{
	m_PV_table.remove (epv);

	ATLTRACE(_T("CServer::Unregister (PV '%s')\n"), epv.getName());
}

// This routine checks if we server a PV of given name.
// The next two conform to the cas libraries two step approach
// of PV creation.
CEpicsPV *CServer::findPV (const char *PVName)
{
    CEpicsPV *pv = m_PV_table.lookup (PVName);

    if (pv!=0  && pv->isDefined())
        return pv;

	return 0;
}

// pvExistTest()
//
// example return from this procedure:
// return pverExistsHere;	// server has PV
// return pverDoesNotExistHere;	// server does know of this PV
// return pverAsynchCompletion;	// deferred result 
pvExistReturn CServer::pvExistTest (const casCtx &ctx, const char *PVName)
{
	CEpicsPV *epv = findPV (PVName);

	//ATLTRACE (_T("CServer::pvExistTest('%s') - %s\n"),
	//	PVName, (epv) ? "Positive" : "Negative");

	return (epv) ? pverExistsHere : pverDoesNotExistHere;
}

// createPV() is called _every_ time that a PV is attached to
// by a client. The name supplied here may be a PV canonical
// (base) name or it may instead be a PV alias name.
//
// example return from this procedure:
// return pPV;	// success (pass by pointer)
// return PV;	// success (pass by ref)
// return S_casApp_pvNotFound; // no PV by that name here
// return S_casApp_noMemory; // no resource to create pv
// return S_casApp_asyncCompletion; // deferred completion
pvCreateReturn CServer::createPV (const casCtx &ctx,	const char *PVName)
{
	CEpicsPV *epv = findPV (PVName);

	ATLTRACE (_T("CServer::createPV('%s') - %s\n"),
		PVName, (epv) ? "OK" : "Error!");

	if (epv)
		return *epv;
	else
		return S_casApp_pvNotFound;
}

// EOF Server.cpp