// Server.h: interface for the CServer class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_SERVER_H__0E17D829_E06E_11D2_8F16_00105AC8D47C__INCLUDED_)
#define AFX_SERVER_H__0E17D829_E06E_11D2_8F16_00105AC8D47C__INCLUDED_

#include <casdef.h>

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "EpicsPV.h"
//class CEpicsPV;
class CServer : public caServer
{
public:
	CServer (size_t hash_table_size);
	virtual ~CServer ();

	// Tell server to serve this PV
	void Register (CEpicsPV &epv);

	// Tell server not to serve this PV any longer
	void Unregister (CEpicsPV &epv);

    CEpicsPV *CServer::findPV (const char *PVName);

	// pvExistTest()
	//
	// example return from this procedure:
	// return pverExistsHere;	// server has PV
	// return pverDoesNotExistHere;	// server does know of this PV
	// return pverAsynchCompletion;	// deferred result 
	virtual pvExistReturn pvExistTest (const casCtx &ctx, 
		const char *pPVAliasName);

	// createPV() is called _every_ time that a PV is attached to
	// by a client. The name supplied here may be a PV canonical
	// (base) name or it may instead be a PV alias name.
	//
	// example return from this procedure:
	// return pPV;	// success (pass by pointer)
	// return PV;	// success (pass by ref)
	// return S_casApp_pvNotFound; // no PV by that name here
	// return S_casApp_noMemory; // no resource to create pv
	// return S_casApp_asyncCompletion; // deferred completion
	virtual pvCreateReturn createPV (const casCtx &ctx,
		const char *pPVAliasName);

private:
	// Hashtable containing all the served PVs:
	resTable<CEpicsPV, stringId>	m_PV_table;
};

#endif // !defined(AFX_SERVER_H__0E17D829_E06E_11D2_8F16_00105AC8D47C__INCLUDED_)

